
# STaPs'17
# WS Corpus analysis, 23 April 2020
# Nils Schiborr, University of Bamberg
# --------------------------------------------------------------------------- #
# EXAMPLE B -- Anaphoric distances in Multi-CAST (v2)


###############################################################################
# --------------------------------------------------------------------------- #
###############################################################################
# BLOCK B1 -- loading packages and sourcing scripts

library(data.table)
library(ggplot2)
library(multicastR)



###############################################################################
# --------------------------------------------------------------------------- #
###############################################################################
# BLOCK B2 -- loading raw data

# download the Multi-CAST (Haig & Schnell 2015) corpus data from the servers
# of the University of Bamberg (~8MB)
mc <- multicast(2101)

# turn the data frame into a data table
mc <- data.table(mc)

# OPTIONAL: look at the data
mc

# column legend:
#  1   corpus    name of the corpus (= language)
#  2   text      name of the text (unique within all corpora)
#  3   uid       utterance ID (unique within a text)
#  4   gword     grammatical words in the object language
#  5   gloss     interlinear morphological glosses (per LGR)
#  6   graid     annotations of form of expression, person/animacy class, and
#                syntactic function with the GRAID scheme (Haig & Schnell 2014)
#  7   gform     extracted GRAID form annotations
#  8   ganim     extracted GRAID person/animacy annotations
#  9   gfunc     extracted GRAID function annotations
# 10   refind    referent indices per the RefIND scheme (Schiborr et al. 2018)
#                (unique within each text)
# 11   isnref    information status of newly introduced referents
#                (see Schiborr et al. 2018: 15)

# OPTIONAL: calculate corpus statistics
mc_clauses(mc)



###############################################################################
# --------------------------------------------------------------------------- #
###############################################################################
# BLOCK B3 -- indexing clauses

# index left clause boundary markers (annotated <#> or <##> on the 'graid'
# column) sequentially by text
mc[grepl("#", graid), cID := seq(1, .N, 1), by = "text"]

# transfer the clause indices to the constituents of each clause
mc[, cID := cID[which(!is.na(cID))], by = cumsum(!is.na(cID))]



###############################################################################
# --------------------------------------------------------------------------- #
###############################################################################
# BLOCK B4 -- calculating anaphoric distances

# remember row numbers to preserve the order of the table
mc[, I := .I]

# create a table that lists the occurrences of every referent index in each
# text alongside its clause index (cID) and row number (I)
refs <- mc[refind != "", .(cID, I), by = c("text", "refind")]

# sort this table by row number
setkey(refs, I)

# calculate distances between co-referential mentions in clause units:
# for each row of the table, subtract its clause index (cID) from the clause
# index of the previous row with the same referent index (refind) (i.e. its
# immediate antecedent) in the same text;
# the first mention of a referent has no antecedent, and so no distance can
# be calculated (i.e. yielding 'NA')
refs[, distance := cID - shift(cID, n = 1, type = "lag"), 
     by = c("text", "refind")]

# join the distance values back into the full table;
# rows without referent indices have no distance value ('NA')
mc <- merge(mc,
            refs[, c("I", "distance")],
            by = "I",
            all.x = TRUE)

# remove working columns
mc[, c("cID", "I") := NULL]



###############################################################################
# --------------------------------------------------------------------------- #
###############################################################################
# BLOCK B5 -- classifying referring expressions and syntactic functions

# if an expression has a GRAID form annotation containing <np>, it is a NP with
# a lexical head (i.e. a common noun or proper name);
# it is a pronominal NP if annotated <pro>, and a zero anaphor if <0>;
# we ignore all other form types (e.g. headless relative clauses)
mc[grepl("np$", gform), expression := "lexical NPs"]
mc[grepl("pro$", gform), expression := "pronominal NPs"]
mc[grepl("(^|_)?0$", gform), expression := "zero anaphors"]
mc[, expression := factor(expression, levels = c("zero anaphors",
                                                 "pronominal NPs",
                                                 "lexical NPs"))]

# we also mark whether an expression is the subject of its clause, that is
# whether it is annotated <:s>, <:a>, or <:ncs> in GRAID (respectively: the
# subjects of intransitive and transitive clauses, and non-canonical subjects
# such as those with quirky case)
mc[, subject := FALSE]
mc[grepl("^(s|a|ncs)(_|$)?", gfunc), subject := TRUE]



###############################################################################
# --------------------------------------------------------------------------- #
###############################################################################
# BLOCK B6 -- calculating proportions

# create a subset with only rows that
# 1. contain lexical or pronominal NPs or zero anaphors
# 2. that are subjects and
# 3. have an anaphoric distance value (i.e. no newly introduced referents)
tmp <- mc[!is.na(expression) & subject == TRUE & !is.na(distance), ]

# OPTIONAL: rein in extreme distance values by treating them as occuring at
# some lower level (winsorization),
# or alternatively just remove them altogether
distLimit <- 5 # can be adjusted!
message(paste0(tmp[distance > distLimit, .N, ], " subject anaphors (",
               round(tmp[distance > distLimit, .N, ] * 100 / nrow(tmp), 0),
               "%) have distances > ", distLimit, " clauses!"))
tmp[distance > distLimit, distance := distLimit]
# tmp <- tmp[distance <= distLimit, ]

# OPTIONAL: exclude anaphors with a distance of zero (i.e. with antecedents in
# the same clause), as these constitute a special case
tmp <- tmp[distance > 0, ]

# calculate proportions of expression types by distance to the antecedent
dist <- CJ(corpus = unique(tmp$corpus),
           distance = unique(tmp$distance),
           expression = unique(tmp$expression))
dist <- merge(x = dist,
              y = tmp[, .N, by = c("corpus", "distance", "expression")],
              by = c("corpus", "distance", "expression"),
              all.x = TRUE)
dist <- merge(x = dist,
              y = tmp[, .(All = .N), by = c("corpus", "distance")],
              by = c("corpus", "distance"),
              all.x = TRUE)
dist[is.na(N), N := 0]
dist[, P := N / All]

# OPTIONAL: look at the table of proportions
dist



###############################################################################
# --------------------------------------------------------------------------- #
###############################################################################
# BLOCK B7 -- plotting the proportions

# create the plot
mcPlot <- ggplot(data = dist,
                 aes(x = factor(distance),
                     y = P,
                     fill = expression)) +
          # split expressions across multiple plots
        #  facet_wrap(~ expression) +
          # draw boxplots
          stat_boxplot(geom = "errorbar",
                       width = 0.4,
                       position = position_dodge(width = 0.7)) +
          geom_boxplot(width = 0.7) +
          # add a plot title
          ggtitle(paste0("Proportions of expressions used for subject ",
                         "anaphors  in ", dist[, .N, by = "corpus"][, .N, ],
                         " Multi-CAST corpora")) +
          # set axis labels
          scale_x_discrete(name = "anaphoric distance in clause units",
                           labels = c(seq(1, distLimit - 1, 1),
                                      paste0(distLimit, "+"))) +
          scale_y_continuous(name = "% of subjects",
                             limits = c(0, 1),
                             breaks = seq(0, 1, 0.2),
                             labels = seq(0, 100, 20)) +
          # set colour scale
          scale_fill_manual(values = c("#D4E9F0", "#7FBCD3", "#2A90B5")) +
          # reposition legend
          theme_bw() +
          theme(legend.title = element_blank(),
                legend.position = "bottom")
mcPlot
