
# calculate the lengths of lexical NP arguments in a UD file
CalculateNPLengths <- function(file) {
  
  # deserialize UD data
  ud <- readRDS(file)
  
  message(paste0(rep("-", 70), collapse = ""))
  message(paste0("Calculating values for file '",
                 ud[1, corpus], "_", ud[1, file], "'"))
  
  # remove unneeded columns
  ud <- ud[, c(1:5, 8, 11:12)]
  
  
  # ------------------------------------------------------------------------- #
  
  # mark rows containing the heads of
  # 1. common lexical NPs (i.e. that have the PoS 'NOUN')
  # 2. that are subjects or (in)direct objects (i.e. that have a dependency
  # relation to their head that is either 'nsubj', 'obj', or 'iobj')
  ud[, lexArg := FALSE]
  ud[pos == "NOUN" & grepl("^(nsubj|obj|iobj)", rType), lexArg := TRUE]
  
  # sequentially index these NPs within each sentence
  ud[lexArg == TRUE, nID := seq(1, .N, 1), by = "sID"]
  
  
  # ------------------------------------------------------------------------- #
  
  # repeatedly self-join the table on dependency relations until all
  # dependencies of lexical NPs are accounted for (i.e. no more links between
  # heads and dependents are left to be made, i.e. links == 0)
  links <- 1
  cycle <- 1
  
  while (links > 0) {
    # self-join tables
    ud <- merge(x = ud,
                y = ud[, c("sID", "wID", "nID")],
                by.x = c("sID", "rID"),
                by.y = c("sID", "wID"),
                suffixes = c("", "_of_head"),
                sort = FALSE,
                all.x = TRUE)
    
    # calculate the number of links made this cycle
    links <- ud[is.na(nID) & !is.na(nID_of_head), .N]
    message(paste0("cycle = ", cycle, ": N = ", links, 
                   " dependencies associated"))
    
    # copy over the head index of the dependency head
    ud[is.na(nID) & !is.na(nID_of_head), nID := nID_of_head]
    
    # remove working column
    ud[, nID_of_head := NULL]
    
    # increment cycle counter
    cycle <- cycle + 1
  }

  
  # ------------------------------------------------------------------------- #
  
  # remove punctuation marks and interjections
  ud <- ud[!grepl("^(PUNCT|INTJ)$", pos), ]
  
  # OPTIONAL: remove elements of multi-word expressions (MWEs)
  # which can be counted as one 'word' in terms of complexity
  ud <- ud[!grepl("^(flat|fixed|compound)$", rType), ]
  
  
  # ------------------------------------------------------------------------- #  
  
  # calculate the length of NPs by counting the number of dependencies linked
  # to each NP (i.e. nID) within each sentence (i.e. sID)
  lengths <- ud[, .(length = .N), by = c("sID", "nID")]
  
  # merge NP lengths into larger table
  ud <- merge(x = ud,
              y = lengths,
              by = c("sID", "nID"),
              all.x = TRUE)
  
  
  # ------------------------------------------------------------------------- #  
  
  # reduce table to just NP heads
  ud <- ud[lexArg == TRUE, ]
  
  # remove working columns
  ud[, c("lexArg", "pos", "nID") := NULL]
  
  # reorder columns
  setcolorder(ud, c("language", "corpus", "file",
                    "sID", "wID", "rID", "rType"))
                    
  # return table
  message("")
  return(ud)
}
